# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import  numpy                                                       as  np
from    sys      import  path                                       as  sys_path
from    os.path  import  abspath, dirname, basename, isfile, isdir
from    os.path  import  join                                       as  pjoin
from    os       import  mkdir                                      as  os_mkdir
from    os       import  listdir                                    as  os_listdir
from    os       import  system                                     as  os_system
from    pprint   import  pformat
from    copy     import  deepcopy
import  torch

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

sys_path.append(dirname(abspath(__file__)))
from  engine_geom_vars  import  mk_geom_vars, get_interpolation_coefficients
sys_path.pop()

# ------------------------------------------------------------------------
#                  Basic Functions
# ------------------------------------------------------------------------

global_stack  =  []

lmap  =  lambda f,x:  list(map(f,x))
lzip  =  lambda  *x:  list(zip(*x))

def splitter(x, L=60):
    A = pformat(x, width=L).replace('"','').replace("'",'').replace("\n",'& \n').split('\n')
    return [c.replace(' ','') for c in A]

coeffs  =  eval(get_interpolation_coefficients())

# def get_inds_closest(y_ghost, y_ref, allow_per = True, return_ext = True):
# 
#     nper, L  =  5, len(y_ref)
#     
#     y_ref    =  y_ref.tolist()
#     for _ in range(nper):
#         y_ref    =  [2*y_ref[0] - y_ref[1]] + y_ref + [2*y_ref[-1] - y_ref[-2]]
#     
#     inds   =  np.searchsorted(y_ref, y_ghost, side='left')  #  a[i-1] < v <= a[i]
#     assert np.all((inds >= 1)*(inds < len(inds)))
# 
#     inds   -=  nper
# 
#     if not allow_per:
#         assert np.all((inds >= 1)*(inds < L))
#     if return_ext:
#         return [(inds-1)%L, (inds)%L, inds-1, inds]
#     else:
#         return [(inds-1)%L, (inds)%L]

class Mk_Extended:
    def __init__(self, x, nper = 10):
        
        self._nper  =  nper

        x    =  x.tolist()
        for _ in range(nper):
            x    =  [2*x[0] - x[1]] + x + [2*x[-1] - x[-2]]
        
        self._x_ext = x
    
    def __call__(self, i):
        i  =  i + self._nper
        assert 0 <= i < len(self._x_ext)
        return self._x_ext[i]

# ------------------------------------------------------------------------
#                  Trilinear Fit
# ------------------------------------------------------------------------

RESIDUAL_CONTAINED = [0]

def contained(target, A , B):
    inside, alpha_a  =  [], []

    for i in range(len(target)):
        if abs(A[i] - B[i]) > 1e-10:
            coord      =  (target[i] - A[i]) / (B[i] - A[i])
            inside .append( -1e-10 <= coord <= (1+1e-10))
            alpha_a.append(  1 - coord                )
    
    if inside:
        if len(inside) == 2:
            if abs(alpha_a[0])>abs(alpha_a[1]): inside, alpha_a = [inside[0]], [alpha_a[0]]
            else                              : inside, alpha_a = [inside[1]], [alpha_a[1]]
        inside,  =   set(inside)
        if inside:
            RESIDUAL_CONTAINED[0] = max(RESIDUAL_CONTAINED[0], abs(max(alpha_a)-min(alpha_a)))
        # assert (len(alpha_a) >= 1) and abs(max(alpha_a)-min(alpha_a)) <= 1e-4, (alpha_a,inside)
        alpha_a  =  (max(alpha_a) + min(alpha_a))/2
        return inside, alpha_a
    
    else:
        for i,arr in enumerate([A,B]):
            if all(abs(t-x)<1e-10 for t,x in zip(target,arr)):
                return True, 1-i
        return False, np.nan

def search_mid_h(xzy_a, xzy_b, get_h):
    xzy_a       =  list(map(np.copy, xzy_a))
    xzy_b       =  list(map(np.copy, xzy_b))

    temp_xzy_a  =  deepcopy(xzy_a)
    temp_xzy_b  =  deepcopy(xzy_b)

    assert len(xzy_a) == len(xzy_b) == 3

    # find points where "a" is higher above the surface than "b" (should not happen)
    m           =  (xzy_a[2] - get_h(*xzy_a[:2])) > (xzy_b[2] - get_h(*xzy_b[:2])) 

    # reverse these points: (a,b = b,a)
    for i in range(3):
        xzy_a[i][m]  =  temp_xzy_b[i][m]
        xzy_b[i][m]  =  temp_xzy_a[i][m]
    del i

    dx, dz, dy  =  [(xzy_b[i] - xzy_a[i]) for i in range(3)]
    x0, z0, y0  =  xzy_a

    eval_h  =  lambda mid: get_h(dx*mid + x0,
                                 dz*mid + z0) - (dy*mid + y0) # positive when below surf (closer to "a")
    
    # define edge case, dx=dz=0
    # in such cases, (y0 = h) and (dy = 0) make code exit at once with right value (search is meaningless)
    m       =  (np.fabs(dx) < 1e-10) * (np.fabs(dz) < 1e-10)
    y0[m]   =  (eval_h(0.5) + (dy*0.5 + y0))[m]
    dy[m]   =  0.

    assert np.all(eval_h(1) <= (eval_h(0)+1e-10))
    assert np.all(         (xzy_b[2] - get_h(*xzy_b[:2])) >= -1e-10) # "b" is above surface
    assert np.all(1e-10 >= (xzy_a[2] - get_h(*xzy_a[:2]))          ) # "a" is below surface

    lo, hi  =  np.zeros_like(dx), np.ones_like(dx)
    iters   =  0

    while np.any((hi-lo) > 1e-10):
        
        iters += 1
        assert iters < 10_000, (lo, hi, xzy_a, xzy_b)

        mid   = (lo + hi) / 2
        new_h =  eval_h(mid)

        m     = (new_h-1e-10) <= 0 # above surf, near b=hi
        hi[m] = mid[m]

        m     = (new_h+1e-10) >= 0 # below surf, near a=lo
        lo[m] = mid[m]

    mid  =  (lo + hi) / 2

    return ( dx*mid + x0,
             dz*mid + z0,
             dy*mid + y0) # eval_h is supposed to be zero (or close)

def apply_Fadlun(fix_points, side, get_h, force_interior = False):
    global_stack.clear() ; global_stack.extend([fix_points, side, get_h, force_interior])
    print('Note: applying Ghost-version of Fadlun IBM scheme!')
    all_args    =   [None for _ in fix_points[side]]
    all_coeffs  =  [[None for _ in range(2)] for _ in fix_points[side]]
    all_inds    =  [[None for _ in range(2)] for _ in fix_points[side]]
    
    all_xn, all_zn, all_yn  =  search_mid_h([[d[f"{c}2"]               for d in fix_points[side]] for c in 'xzy'], 
                                            [[d['fluid_points'][-1][c] for d in fix_points[side]] for c in 'xzy'],
                                            get_h                                                                )
    RESIDUAL_CONTAINED[0] = 0
    for ind in range(len(all_inds)):
        d                 =  fix_points[side][ind]

        ijk_pi  =  [[arr[c] for c in 'ijk'] for arr in d['fluid_points']]
        xzy_pi  =  [[arr[c] for c in 'xzy'] for arr in d['fluid_points']]

        xn,zn,yn = xzy_n  =  all_xn[ind], all_zn[ind], all_yn[ind]
        xzy_g             =  [2*xn - d['x2'],
                              2*zn - d['z2'],
                              2*yn - d['y2']]
        
        xzy_p1, xzy_p2  =  xzy_pi[:2]
        ijk_p1, ijk_p2  =  ijk_pi[:2]

        contain_n1, alpha_n   =  contained(xzy_g, xzy_n , xzy_p1)
        contain_12, alpha_p1  =  contained(xzy_g, xzy_p1, xzy_p2)

        # assert (contain_n1 + contain_12) >= 1, (alpha_n, alpha_p1, xzy_g, xzy_n , xzy_p1, xzy_p2, side, ind, d)
        if contain_n1 and contain_12:
            assert abs(alpha_n)<1e-10 and abs(alpha_p1-1)<1e-10
            
        all_coeffs[ind] = [0.     for _ in all_coeffs[ind]]
        all_inds  [ind] = [ijk_p1 for _ in all_inds  [ind]]

        if contain_n1 and (not force_interior):
            # use point, if contained here, and not forcing interior
            #     -> forced point is used later in first triplet (ip,ip+1)
            # all_inds  [ind][0]  =    ijk_p1  # done before
            all_coeffs[ind][0]  =  -(1-alpha_n) # negative signs
        else:
            for ip in range(len(ijk_pi)-1):
                contain__, alpha___  =  contained(xzy_g, xzy_pi[ip], xzy_pi[ip+1])
                if contain_n1 and force_interior:
                    contain__ = True # force previous choice
                if contain__:
                    all_inds  [ind][0]  =    ijk_pi[ip]
                    all_coeffs[ind][0]  =  -(    alpha___) # negative signs
                    all_inds  [ind][1]  =          ijk_pi[ip+1]
                    all_coeffs[ind][1]  =  -(1 - alpha___)
                    break
            assert contain__
        
        all_args[ind] = {}
    if RESIDUAL_CONTAINED[0] > 1e-6:
        print(f'WARNING: max. residual in alphas from contained function was {RESIDUAL_CONTAINED[0]:.6e} {side = }')
    return all_inds, all_coeffs, all_args

# ------------------------------------------------------------------------
#                 Make RUVWT
# ------------------------------------------------------------------------
def mk_Ru():
    i0   =  'i-1'
    i2   =  'i+1'
    j0   =  'j-1'
    j2   =  'j+1'

    k0   =  'k-1'
    k2   =  'k+1'

    uu   =  'U(i,j,k)'

    vv   =  f'0.25*(V(i,j,k)   + V({i0},j,k) +'    \
            f'      V(i,j,k+1) + V({i0},j,k+1))'
            
    ww   =  f'0.25*(W(i,j,k)    + W({i0},j,k) +'    \
            f'      W(i,{j2},k) + W({i0},{j2},k))'

    ddy   =  f'U(i,j,{k0}) * Ruw_ddy_a(k)+'  \
             f'U(i,j,k)    * Ruw_ddy_b(k)+'  \
             f'U(i,j,{k2}) * Ruw_ddy_c(k)'      
            
    d2dy2 =  f'U(i,j,{k0}) * Ruw_d2dy2_a(k)+'  \
             f'U(i,j,k)    * Ruw_d2dy2_b(k)+'  \
             f'U(i,j,{k2}) * Ruw_d2dy2_c(k)'      
            
    ddx   = f'(U({i2},j,k) - U({i0},j,k))*inv_dx__2'  
    ddz   = f'(U(i,{j2},k) - U(i,{j0},k))*inv_dz__2'  

    d2dx2 = f'(U({i2},j,k) -2*{uu} + U({i0},j,k))*inv_dx2'
    d2dz2 = f'(U(i,{j2},k) -2*{uu} + U(i,{j0},k))*inv_dz2'

    expl_term      =  f'-({uu})*({ddx}) -({vv})*({ddy}) -({ww})*({ddz}) + nu*({d2dx2} + {d2dy2} + {d2dz2}) + Sf_x'
    return splitter(expl_term)

def mk_Rv():
    i0   =  'i-1'
    i2   =  'i+1'
    j0   =  'j-1'
    j2   =  'j+1'

    k0   =  'k-1'
    k2   =  'k+1'

    kk   =  'k'
    kk0  =  'k-1'

    uu  =  f'0.5*(U(i,j,{kk0}) + U({i2},j,{kk0}))*Rv_interp_uw_bottom(k) +' \
           f'0.5*(U(i,j,{kk})  + U({i2},j,{kk}))*Rv_interp_uw_top(k)'
            
    vv  =  'V(i,j,k)'

    ww  =  f'0.5*(W(i,j,{kk0}) + W(i,{j2},{kk0}))*Rv_interp_uw_bottom(k) +' \
           f'0.5*(W(i,j,{kk})  + W(i,{j2},{kk}))*Rv_interp_uw_top(k)'

    ddy   =  f'V(i,j,{k0}) * Rv_ddy_a(k) +' \
             f'V(i,j,k)    * Rv_ddy_b(k) +' \
             f'V(i,j,{k2}) * Rv_ddy_c(k)'
            
    d2dy2 =  f'V(i,j,{k0}) * Rv_d2dy2_a(k) +' \
             f'V(i,j,k)    * Rv_d2dy2_b(k) +' \
             f'V(i,j,{k2}) * Rv_d2dy2_c(k)'
            
    ddx   = f'(V({i2},j,k) - V({i0},j,k))*inv_dx__2'
    ddz   = f'(V(i,{j2},k) - V(i,{j0},k))*inv_dz__2'

    d2dx2 = f'(V({i2},j,k) -2*{vv} + V({i0},j,k))*inv_dx2'
    d2dz2 = f'(V(i,{j2},k) -2*{vv} + V(i,{j0},k))*inv_dz2'

    expl_term  =  f'-({uu})*({ddx}) - ({vv})*({ddy}) - ({ww})*({ddz}) + nu*({d2dx2} + {d2dy2} + {d2dz2})'
    return splitter(expl_term)

def mk_Rw():
    
    i0   =  'i-1'
    i2   =  'i+1'
    j0   =  'j-1'
    j2   =  'j+1'

    k0  =  'k-1'
    k2  =  'k+1'

    uu  =  f'0.25*(U(i,j,k)    + U(i,{j0},k) +' \
           f'      U({i2},j,k) + U({i2},{j0},k))'

    vv  =  f'0.25*(V(i,j,k)   + V(i,{j0},k) +' \
           f'      V(i,j,k+1) + V(i,{j0},k+1))'

    ww  =   'W(i,j,k)'
    

    ddy =  f'W(i,j,{k0}) * Ruw_ddy_a(k) +' \
           f'W(i,j,k)    * Ruw_ddy_b(k) +' \
           f'W(i,j,{k2}) * Ruw_ddy_c(k)'
    
    d2dy2 =  f'W(i,j,{k0}) * Ruw_d2dy2_a(k) +' \
             f'W(i,j,k)    * Ruw_d2dy2_b(k) +' \
             f'W(i,j,{k2}) * Ruw_d2dy2_c(k)'
    
    ddx   = f'(W({i2},j,k) - W({i0},j,k))*inv_dx__2'
    ddz   = f'(W(i,{j2},k) - W(i,{j0},k))*inv_dz__2'

    d2dx2 = f'(W({i2},j,k) -2*{ww} + W({i0},j,k))*inv_dx2'
    d2dz2 = f'(W(i,{j2},k) -2*{ww} + W(i,{j0},k))*inv_dz2'

    expl_term     =  f'-({uu})*({ddx}) - ({vv})*({ddy}) - ({ww})*({ddz}) + nu*({d2dx2} + {d2dy2} + {d2dz2}) + Sf_z'
    return splitter(expl_term)

def mk_Rt():
    i0   =  'i-1'
    i2   =  'i+1'
    j0   =  'j-1'
    j2   =  'j+1'

    k0   = 'k-1'
    k2   = 'k+1'

    uu   =   f'0.5*(U(i,j,k) + U({i2},j,k))'
    vv   =   f'0.5*(V(i,j,k) + V(i,j,k+1))'
    ww   =   f'0.5*(W(i,j,k) + W(i,{j2},k))'

    ddy =  f'T(i,j,{k0}) * Ruw_ddy_a(k) +' \
           f'T(i,j,k)    * Ruw_ddy_b(k) +' \
           f'T(i,j,{k2}) * Ruw_ddy_c(k)'
    
    ddx   = f'(T({i2},j,k) - T({i0},j,k))*inv_dx__2'
    ddz   = f'(T(i,{j2},k) - T(i,{j0},k))*inv_dz__2'

    d2dy2 =  f'T(i,j,{k0}) * Ruw_d2dy2_a(k) +' \
             f'T(i,j,k)    * Ruw_d2dy2_b(k) +' \
             f'T(i,j,{k2}) * Ruw_d2dy2_c(k)'

    d2dx2 = f'(T({i2},j ,k) -2*T(i,j,k) + T({i0},j ,k))*inv_dx2'
    d2dz2 = f'(T(i ,{j2},k) -2*T(i,j,k) + T(i ,{j0},k))*inv_dz2'

    expl_term        = f'-({uu})*({ddx}) - ({vv})*({ddy}) - ({ww})*({ddz}) + cond*({d2dx2} + {d2dy2} + {d2dz2}) + Sq'
    return splitter(expl_term)

# ------------------------------------------------------------------------
#                 Fill Geom_Vars
# ------------------------------------------------------------------------

def extra_cells(y,n):
    if hasattr(y,'shape'):
        y = y.reshape(-1).tolist()
    for _ in range(n):
        y = [2*y[0] - y[1]] + y + [2*y[-1] - y[-2]]
    return y

# P is not affected, so this is good
def mk_dPdy(yu, yp, yp_extended):
    gv  =  mk_geom_vars(yu[:-1], 'neumann', y_base=yp_extended, order=3)
    D   =  {}
    for c in 'abcd':
        D[f"dPdy_ind_{c}"] = inds  =  gv[f'inds_{c}']
        D[f"dPdy_ddy_{c}"] = ddy   =  gv[f'ddy_{c}']
        for j in range(len(inds)):
            inds[j]     -= 1
            
            while inds[j] < 0       :
                inds[j] += 1
                assert abs(ddy[j]) < 1e-10
            
            while inds[j] >= len(yp):
                inds[j] -= 1
                assert abs(ddy[j]) < 1e-10
    return D

def fill_geom_vars(yu, yp, yp_extended, ny, nx_nz, geom_vars_folder):  
    assert len(yu)          == (ny+1)
    assert len(yp)          ==  ny
    assert len(yp_extended) == (ny+2)
    assert yu[ 0] == yp_extended[ 0]
    assert yu[-1] == yp_extended[-1]

    assert yu[0] < yp[0] < yp[-1] < yu[-1]
    
    def fwrite(tag, values):
        fname  =  pjoin(geom_vars_folder, f"{tag}.dat")
        assert isdir(dirname(fname)) and (not isfile(fname))
        with open(fname, 'w') as f:
            for val in values:
                f.write(f"{val}\n")

    fwrite('inv_dy_rhs', (1./np.diff(yu)).tolist())
    
    
    def gen_Rv_interp():
        Rv_interp_uw_top    = []
        Rv_interp_uw_bottom = []
        n_passed = 0
        for i,yyu in enumerate(yu):
            if i in [0,ny]:
                Rv_interp_uw_top   .append(0.)
                Rv_interp_uw_bottom.append(0.)
                n_passed += 1
            else:
                dy_top     =  yp[i] - yyu
                dy_bottom  =          yyu - yp[i-1]
                Rv_interp_uw_top   .append( 1 - (dy_top   /(dy_top + dy_bottom)) )
                Rv_interp_uw_bottom.append( 1 - (dy_bottom/(dy_top + dy_bottom)) )
        fwrite('Rv_interp_uw_top'   , Rv_interp_uw_top   )
        fwrite('Rv_interp_uw_bottom', Rv_interp_uw_bottom)
        assert n_passed == 2
    gen_Rv_interp()

    # P_band_a            
    # P_band_b            
    # P_band_c 

    # can read ghost, and give it no importance
    def mk_P_bands():
        gv            =  mk_geom_vars(extra_cells(yp,1), type_bc='normal')
        a,b,c         =  [gv[f'd2dy2_{c}'][1:-1] for c in 'abc']
        b[-1]        += c[-1]
        b00           =  [val for val in b]
        b  [0]       += a[0]
        b00[0]       -= a[0]
        a[0] = c[-1]  = 0

        fwrite(f'P_band_a'   , a  )
        fwrite(f'P_band_a_00', a  )

        fwrite(f'P_band_b'   , b  )
        fwrite(f'P_band_b_00', b00)

        fwrite(f'P_band_c'   , c  )
        fwrite(f'P_band_c_00', c  )

    mk_P_bands()

    # Ruw_inds_ia         
    # Ruw_inds_ic   
    # Ruw_ddy_a           
    # Ruw_ddy_b           
    # Ruw_ddy_c           
    # Ruw_d2dy2_a         
    # Ruw_d2dy2_b         
    # Ruw_d2dy2_c  

    # can read ghost, and give it no importance
    def mk_Ruw_derivs():
        gv  =  mk_geom_vars(yp_extended, type_bc='dirichlet')
        for c in 'abc':
            fwrite(f'Ruw_ddy_{c}'      , gv[f'ddy_{c}']  [1:-1])
            fwrite(f'Ruw_d2dy2_{c}'    , gv[f'd2dy2_{c}'][1:-1])
    mk_Ruw_derivs()

    # Rv_inds_ia          
    # Rv_inds_ic   
    # Rv_ddy_a            
    # Rv_ddy_b            
    # Rv_ddy_c            
    # Rv_d2dy2_a          
    # Rv_d2dy2_b          
    # Rv_d2dy2_c

    # V will have normal derivatives, reading ghost cell
    def mk_Rv_derivs():
        gv  =  mk_geom_vars(yu, type_bc='dirichlet')
        for c in 'abc':
            fwrite(   f'Rv_ddy_{c}'             , gv[f'ddy_{c}'  ])
            fwrite(   f'Rv_d2dy2_{c}'           , gv[f'd2dy2_{c}'])
    mk_Rv_derivs()

    # dPdy_ddy_a 
    # dPdy_ddy_b 
    # dPdy_ddy_c 
    # dPdy_ddy_d
    
    for tag, entry in  mk_dPdy(yu, yp, yp_extended).items():
        fwrite(tag, entry)
    del tag, entry

    fwrite('yp_coord', yp.tolist())
    fwrite('yu_coord', yu.tolist())


# ------------------------------------------------------------------------
#                 Make rough_surface
# ------------------------------------------------------------------------

def make_rough_surface(info, new_folder):
    xp, xu  =  info.xp, info.xu
    zp, zu  =  info.zp, info.zu
    yp, yu  =  info.yp, info.yu
    for tag, x, z, y in [['T', xp, zp, yp],
                         ['U', xu, zp, yp],
                         ['V', xp, zp, yu],
                         ['W', xp, zu, yp]]: # the indexing of upp matches fortran, because yu[0]==0 in fortran too
        x, z, y         = lmap(np.array,[x, z, y])
        Hb              =  info.get_h(x.reshape(-1,1), z.reshape(1,-1), side = 'bottom')
        Ht              =  info.get_h(x.reshape(-1,1), z.reshape(1,-1), side = 'top'   )

        # print(f"{Hb.max()                              = }")
        # print(f"{Hb.min()                              = }")
        # print(f"{y.min()                               = }")
        # print(f"{Hb.min()-y.min()                      = }")
        # print(f"{min(np.diff(y))                       = }")
        # print(f"{abs(Hb.min()-y.min())/min(np.diff(y)) = }")

        assert y.tolist() == sorted(y)
        assert y.min()     < Hb.min()

        # if Hb.min() < y.min(): assert abs(Hb.min()-y.min()) < 0.5*min(np.fabs(np.diff(y)))
        
        # assert Hb.max() < (info.Ly/3)

        inds            =  {}
        inds['bottom']  =  np.searchsorted(y, Hb, side='left' )      # left: a[i-1]  <  v <= a[i]    # H <= y[i]
        inds['top']     =  np.searchsorted(y, Ht, side='right') - 1  # right: a[i-1] <= v <  a[i]    #            y[i-1] <= H

        assert np.all((0<inds['bottom'])*(inds['bottom']<(len(y)-1)))
        assert np.all((0<inds['top']   )*(inds['top']   <(len(y)-1)))
        
        for side in ['bottom','top']:
            ic = inds[side]
            y0 = {'bottom': Hb, 'top': Ht}[side]
            d  = {'bottom':  1, 'top': -1}[side]
            yc = y[ic]
            y1 = y[ic +   d]
            y2 = y[ic + 2*d]
            A_write = []
            for p in '12':
                val = coeffs['2']['dirichlet_0']['f'][f"p{p}"](yc, y0, y1, y2)
                A_write.append(val)
            coeffs_b, coeffs_c = A_write
            short_side  =  {'bottom':  'bot', 'top': 'top'}[side]
            with open(pjoin(new_folder,f'roughsurf_{tag}_{short_side}_ij_ic_coeffs_bc.dat'),'w') as f:
                for     i in range(A_write[0].shape[0]):
                    for j in range(A_write[0].shape[1]):
                        f.write(f"{i} {j} {ic[i,j]} {coeffs_b[i,j]} {coeffs_c[i,j]}" + '\n')

# ------------------------------------------------------------------------
#                 Make rough_surface
# ------------------------------------------------------------------------

def mk_d_xzy(mult_distance):
    result = [[dx,dz,dy] for dx in [0, -1, 1]
                         for dz in [0, -1, 1]
                         for dy in [0, -1, 1]]
    result.sort(key = lambda val: [sum(map(abs,val)), -abs(val[2]),-abs(val[0]), -abs(val[1]) ])
    assert result[0] == [0,0,0]
    result =  np.array(result[1:])
    result2 = []
    for mm in sorted(mult_distance):
        result2.extend((mm*result).tolist())
    return result2

def make_rough_surface_ghost(info, new_folder, temp_output = False, return_variables = False,
                             mult_distance = [1], force_interior = False, use_tags = 'UVWT'):
    extra_holders  =  {}
    if return_variables: final_results = {}
    for tag, x, z, y in [['T', info.xp, info.zp, info.yp],
                         ['U', info.xu, info.zp, info.yp],
                         ['V', info.xp, info.zp, info.yu],
                         ['W', info.xp, info.zu, info.yp]]: # the indexing of upp matches fortran, because yu[0]==0 in fortran too
        if not tag in use_tags: continue
        x, z, y         =  lmap(lambda arr: np.array(arr).reshape(-1),[x, z, y])
        all_H           =  {'Hb':  info.get_h(x.reshape(-1,1), z.reshape(1,-1), side = 'bottom'),
                            'Ht':  info.get_h(x.reshape(-1,1), z.reshape(1,-1), side = 'top'   )}
        x_ext           =  Mk_Extended(x)
        z_ext           =  Mk_Extended(z)
        # a) find points to be fixed
        def get_fix_points():
            fix_points      =  {'top': [], 'bottom': []}
            Hb, Ht          =  all_H['Hb'], all_H['Ht']

            assert y.tolist() == sorted(y)

            is_fluid    =  ((Hb.reshape(*Hb.shape,1) <= y.reshape( 1, 1,-1)                           ) * 
                            (                           y.reshape( 1, 1,-1) <= Ht.reshape(*Ht.shape,1)))
            is_ok       =  np.copy(is_fluid)

            nx, nz, ny  =  is_fluid.shape

            for dx,dz,dy in  mk_d_xzy(mult_distance):
                for          ii in  range(is_fluid.shape[0]):
                    for      jj in  range(is_fluid.shape[1]):
                        for  kk in  range(is_fluid.shape[2]):
                            if is_fluid[ii,jj,kk]:
                                side      =  'bottom' if kk<=(ny//2) else 'top' # rough surface is away from middle anyway, so this is ok
                                i2,j2,k2  =  ( (ii + dx)%nx ,
                                               (jj + dz)%nz ,
                                                kk + dy     )
                                if 0<= k2 < ny:
                                    if not is_ok[i2,j2,k2]:
                                        mm                =  max(map(abs, [dx,dz,dy]))
                                        dx_u, dz_u, dy_u  =  dx//mm, dz//mm, dy//mm
                                        assert ((dx_u*mm) == dx) and ((dz_u*mm) == dz) and ((dy_u*mm) == dy) 
                                        fix_points[side].append(dict(i2           = i2                                                              ,
                                                                     j2           = j2                                                              ,
                                                                     k2           = k2                                                              ,
                                                                     x2           = x_ext(ii + dx)                                                  ,
                                                                     z2           = z_ext(jj + dz)                                                  ,
                                                                     y2           =     y[k2]                                                       ,
                                                                     x_start      = (x_ext(ii) + x_ext(ii + dx))/2                                  ,
                                                                     z_start      = (z_ext(jj) + z_ext(jj + dz))/2                                  ,
                                                                     fluid_points = [dict(i  =      (ii-dx_u*m)%nx, j  =      (jj-dz_u*m)%nz, k  =   kk-dy_u*m ,
                                                                                          x  = x_ext(ii-dx_u*m)   , z  = z_ext(jj-dz_u*m)   , y  = y[kk-dy_u*m]) for m in range(mm+1)],
                                                                    ))
                                        is_ok[i2,j2,k2] = True
            return fix_points
        fix_points = get_fix_points()

        # b) find trilinear interpolation coefficients
        def get_trilinear(fix_points):
            result_inds   = []
            result_coeffs = []
            result_target = []
            result_args   = []
            
            for  side  in  ['top', 'bottom']:
                get_h           =  lambda x,z: info.get_h(x,z, side = side ,
                                                               lib  = np   )
                all_inds, all_coeffs, all_args = apply_Fadlun(fix_points, side, get_h, force_interior = force_interior)

                result_coeffs.extend(all_coeffs)
                result_inds  .extend(all_inds  )
                if temp_output:
                    result_args.extend(all_args)
                result_target.extend([[d[f'{c}2'] for c in 'ijk'] for d in fix_points[side]])
            return result_coeffs, result_inds, result_target, result_args
        result_coeffs, result_inds, result_target, result_args  =  get_trilinear(fix_points)

        if return_variables:
            final_results[tag] = dict(coeffs = result_coeffs,
                                      inds   = result_inds  ,
                                      target = result_target,
                                      args   = result_args  )
        else:
            fname  =  pjoin(new_folder,f'roughsurf_{tag}_ghost_interp.dat')
            with open(fname,'w') as f:
                fmt_ijk  =  lambda i,j,k: f"{i} {j} {k}"
                write    =  lambda x    : f.write(x + '\n')
                if temp_output:
                    # # ------- temp -------
                    write(f"! x = {x.tolist()}")
                    write(f"! z = {z.tolist()}")
                    write(f"! y = {y.tolist()}")
                    # # ------- end temp -------
                for ind in range(len(result_inds)):
                    if temp_output:
                        write(  fmt_ijk(*result_target[ind]) + f' ! {result_args[ind]}')
                    else:
                        write(  fmt_ijk(*result_target[ind])  )  #  + f' ! {result_args[ind]}')
                    for j in range(len(result_inds[ind])):
                        write(f'{fmt_ijk(*result_inds[ind][j])} {result_coeffs[ind][j]}')
            print(f'Done writing coeffs: {basename(fname)} (min: {np.array(result_coeffs).min()}) (max: {np.array(result_coeffs).max()})')
            extra_holders[f"PYTHON_HOLDER_n_rough_{tag}"] = len(result_inds)
    if return_variables:
        return final_results
    else:
        return extra_holders

# ------------------------------------------------------------------------
#                 Direct Runner
# ------------------------------------------------------------------------

if __name__ == '__main__':
    
    print('\n\nRu expression')
    for x in mk_Ru():
        print(x)
    

    print('\n\nRv expression')
    for x in mk_Rv():
        print(x)

    print('\n\nRw expression')
    for x in mk_Rw():
        print(x)
    

    print('\n\nRt expression')
    for x in mk_Rt():
        print(x)